#include "DDHUndoableOrthKnife.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "c3d.h"

//---------------------------------------------------------------------------------------
DDHUndoableOrthKnife::DDHUndoableOrthKnife( ACObject *ko, bool delknife, int view )
{

   mUndoCalled = false;

   // Knife elements
   mKnifeObject = ko;
   mDeleteKnifeInd = delknife;
   mView = view;

   // Undo/redo data
   mDeletedKnife = NULL;
   mDeletedKnifeOwner = NULL;
   mNewVerts = NULL;

   // Ensure the knife object is valid
   if( !object_exists( mKnifeObject ))
   {
      message_dialog( "orth knife : knife is not specified or no longer exists" );
      abort();
      return;
   }

   // Abort if object has children
   if( ac_object_get_childrenlist( mKnifeObject ) != NULL )
   {
      message_dialog( "orth knife : knife must be a single object (with no child objects)" );
      abort();
      return;
   }
   // Abort if knife is not a poly
   if( ac_object_get_type( mKnifeObject ) != OBJECT_NORMAL )
   {
      message_dialog( "orth knife : knife must be a polygonal object" );
      abort();
      return;
   }

}
//---------------------------------------------------------------------------------------
DDHUndoableOrthKnife::~DDHUndoableOrthKnife()
{

   if( mUndoCalled )
   {
      // Remove the new vertices
      for( List *pv = mNewVerts; pv != NULL; pv = pv->next )
      {
         CNewVert *nv = ( CNewVert * )( pv->data );
         vertex_free( nv->v );
         delete nv;
      }
   }
   else
   {
      // Destroy data - the objects are no longer present
      object_free( mDeletedKnife );
   }

   list_free( &mNewVerts );

}
//---------------------------------------------------------------------------------------
DDHUndoableOrthKnife::CNewVert *DDHUndoableOrthKnife::SliceEdge( ACEdge *edge, ACEdge *knifeedge )
{

   // Generate a plane normal based on ortho axis to slice with
   Vertex v3;
   switch( mView )
   {
   case 1:
      // XZ
      v3.x = knifeedge->v1->x;
      v3.y = knifeedge->v1->y + 1;
      v3.z = knifeedge->v1->z;
      break;
   case 2:
      // YZ
      v3.x = knifeedge->v1->x + 1;
      v3.y = knifeedge->v1->y;
      v3.z = knifeedge->v1->z;
      break;
   case 0:
      // XY
      v3.x = knifeedge->v1->x;
      v3.y = knifeedge->v1->y;
      v3.z = knifeedge->v1->z + 1;
      break;
   }

   Vertex edgenorm;
   C3d::CalcFaceNormal( knifeedge->v1, knifeedge->v2, &v3, &edgenorm );

   // Get the intersection of the edge and the knife plane
   Vertex newvertloc;
   int res = C3d::SegmentPlaneIntersect( edge->v1, edge->v2, knifeedge->v1, &edgenorm, &newvertloc );
   if( res != 0 )
   {
      // No intersection - bail
      return NULL;
   }

   // Determine whether new point is in the segment
   switch( mView )
   {
   case 1:
      // XZ
      if( knifeedge->v1->x != knifeedge->v2->x )
      {
         if( knifeedge->v1->x <= newvertloc.x && newvertloc.x <= knifeedge->v2->x )
         {
         }
         else if( knifeedge->v1->x >= newvertloc.x && newvertloc.x >= knifeedge->v2->x )
         {
         }
         else
         {
            return NULL;
         }
      }
      else
      {
         if( knifeedge->v1->z <= newvertloc.z && newvertloc.z <= knifeedge->v2->z )
         {
         }
         else if( knifeedge->v1->z >= newvertloc.z && newvertloc.z >= knifeedge->v2->z )
         {
         }
         else
         {
            return NULL;
         }
      }
      break;
   case 2:
      // YZ
      if( knifeedge->v1->y != knifeedge->v2->y )
      {
         if( knifeedge->v1->y <= newvertloc.y && newvertloc.y <= knifeedge->v2->y )
         {
         }
         else if( knifeedge->v1->y >= newvertloc.y && newvertloc.y >= knifeedge->v2->y )
         {
         }
         else
         {
            return NULL;
         }
      }
      else
      {
         if( knifeedge->v1->z <= newvertloc.z && newvertloc.z <= knifeedge->v2->z )
         {
         }
         else if( knifeedge->v1->z >= newvertloc.z && newvertloc.z >= knifeedge->v2->z )
         {
         }
         else
         {
            return NULL;
         }
      }
      break;
   case 0:
      // XY
      if( knifeedge->v1->x != knifeedge->v2->x )
      {
         if( knifeedge->v1->x <= newvertloc.x && newvertloc.x <= knifeedge->v2->x )
         {
         }
         else if( knifeedge->v1->x >= newvertloc.x && newvertloc.x >= knifeedge->v2->x )
         {
         }
         else
         {
            return NULL;
         }
      }
      else
      {
         if( knifeedge->v1->y <= newvertloc.y && newvertloc.y <= knifeedge->v2->y )
         {
         }
         else if( knifeedge->v1->y >= newvertloc.y && newvertloc.y >= knifeedge->v2->y )
         {
         }
         else
         {
            return NULL;
         }
      }
      break;
   }
   /*
   if( !( C3d::CollinearPointInSegment( &newvertloc, knifeedge->v1, knifeedge->v2 )))
   {
      return NULL;
   }
   */

   // Add the vertex between the edge vertices
   Boolean wasnew;
   CNewVert *nv = new CNewVert();
   nv->edgev1 = edge->v1;
   nv->edgev2 = edge->v2;
   nv->o = object_of_vertex( edge->v1 );
   nv->v = ac_edge_insert_vertex( edge, &wasnew );

   // Get the % distance between the new vertex and its final position
   float newdist = sqrtf( pow( edge->v1->x - newvertloc.x, 2 ) +
                          pow( edge->v1->y - newvertloc.y, 2 ) +
                          pow( edge->v1->z - newvertloc.z, 2 ));
   float origdist = sqrtf( pow( edge->v1->x - edge->v2->x, 2 ) +
                           pow( edge->v1->y - edge->v2->y, 2 ) +
                           pow( edge->v1->z - edge->v2->z, 2 ));
   float pctdist = fabs( newdist ) / fabs( origdist );

   // Adjust the texture coordinates
   for( List *sl = edge->surfacelist; sl != NULL; sl = sl->next )
   {
      Surface *s = ( Surface * )( sl->data );
      SVertex *sthis = surface_get_svertex( s, nv->v );
      SVertex *sfirst = surface_get_svertex( s, edge->v1 );
      SVertex *ssecond = surface_get_svertex( s, edge->v2 );
      sthis->tx = sfirst->tx - (( sfirst->tx - ssecond->tx ) * pctdist );
      sthis->ty = sfirst->ty - (( sfirst->ty - ssecond->ty ) * pctdist );
   }
   nv->v->x = newvertloc.x;
   nv->v->y = newvertloc.y;
   nv->v->z = newvertloc.z;

   // Add the vertex to the list
   list_add_item_head( &mNewVerts, nv );

   return nv;

}
//---------------------------------------------------------------------------------------
void DDHUndoableOrthKnife::DoSlice()
{

   // Grab all of the edges in the current selection
   List *edges = ac_selection_get_edges( false );

   // Do not include edges that are already part of the knife object
   for( List *lp = edges; lp != NULL; lp = lp->next )
   {
      ACEdge *e = ( ACEdge * )( lp->data );
      if( object_of_vertex( e->v1 ) == mKnifeObject )
      {
         message_dialog( "orthknife : the knife object is selected - please deselect it" );
         ac_edgelist_free( &edges );
         abort();
         return;
      }
   }

   // Get edges of knife
   List *knifeedges = ac_object_get_edges( mKnifeObject );

   // Slice each selected edge
   for( List *pe = edges; pe != NULL; pe = pe->next )
   {
      ACEdge *e = ( ACEdge * )( pe->data );
      // Iterate through each knife edge
      for( List *pke = knifeedges; pke != NULL; pke = pke->next )
      {
         ACEdge *ke = ( ACEdge * )( pke->data );
         // Intersect the line segments
         if( SliceEdge( e, ke ) != NULL )
         {
            // Found an intersect
            break;
         }
      }
   }

   // Cleanup the ac3d edges we pulled
   ac_edgelist_free( &edges );
   ac_edgelist_free( &knifeedges );

   selected_calc_normals();

   // Remove the knife
   if( mDeleteKnifeInd )
   {
      mDeletedKnife = mKnifeObject;
      mDeletedKnifeOwner = object_parent( mDeletedKnife );
      ac_unselect_object( mDeletedKnife );
      object_remove_child_nocleanup( mDeletedKnifeOwner, mDeletedKnife );
   }

   redraw_all();
   find_bounding_box();
   display_status();

}
//---------------------------------------------------------------------------------------
void DDHUndoableOrthKnife::execute()
{
   mUndoCalled = false;
   DoSlice();
}
//---------------------------------------------------------------------------------------
void DDHUndoableOrthKnife::undo()
{

   mUndoCalled = true;

   // Recreate the knife
   if( mDeleteKnifeInd )
   {
      object_add_child( mDeletedKnifeOwner, mDeletedKnife );
   }

   // Remove the new vertices
   for( List *pv = mNewVerts; pv != NULL; pv = pv->next )
   {
      CNewVert *nv = ( CNewVert * )pv->data;
      object_delete_vertex( nv->o, nv->v );
      delete nv;
   }

   list_free( &mNewVerts );
   mNewVerts = NULL;

   redraw_all();
   find_bounding_box();

}
//---------------------------------------------------------------------------------------
