#include "DDHUndoableOrthKnife.h"
#include <stdio.h>
#include <string.h>
#include "c3d.h"

// Global knife edges
static ACObject *ddhStaticOrthKnife = NULL;

//---------------------------------------------------------------------------------------
// Set the knife object
// Called from TCL knife interface ("Set knife") to set the object to be used as the knife.
// Sets the static "staticKnifeObject" variable to the currently selected object.
static void ddhorthknife_setKnife()
{

   ACObject *o = NULL;
   List *l = NULL;

   // Get the object
   bool oneobj = true;
   switch( ac_get_select_mode() )
   {
   case SELECT_VERTEX:
      l = ac_selection_get_vertices();
      for( List *lp = l; lp != NULL; lp = lp->next )
      {
         if( o == NULL )
         {
            o = object_of_vertex(( Vertex * )( lp->data ));
         }
         else if( object_of_vertex(( Vertex * )( lp->data )) != o )
         {
            oneobj = false;
            break;
         }
      }
      break;
   case SELECT_SURFACE:
      l = ac_selection_get_surfaces();
      for( List *lp = l; lp != NULL; lp = lp->next )
      {
         if( o == NULL )
         {
            o = object_of_surface(( Surface * )( lp->data ));
         }
         else if( object_of_surface(( Surface * )( lp->data )) != o )
         {
            oneobj = false;
            break;
         }
      }
      break;
   default:
      l = ac_selection_get_objects();
      o = ( ACObject * )( l->data );
      if( list_count( l ) > 1 )
      {
         oneobj = false;
      }
      break;
   }

   if( o == NULL )
   {
      message_dialog( "orth knife : no object selected, or object has no geometry" );
      return;
   }
   else if( !oneobj )
   {
      message_dialog( "orth knife : select only one object" );
      list_free( &l );
      return;
   }

   ddhStaticOrthKnife = o;
   list_free( &l );
   char *kname = ac_object_get_name( ddhStaticOrthKnife );
   if( kname != NULL && strlen( kname ) > 0 )
   {
      tcl_command( "set ddhorthknife_knife_name \"%s\"", kname );
   }
   else
   {
      tcl_command( "set ddhorthknife_knife_name \"Unnamed\"" );
   }

}
//---------------------------------------------------------------------------------------
// Perform a knife cut
static void ddhorthknife_selectedGo( int argc, const char* argv[] )
{

   if( argc < 4 )
   {
      message_dialog( "Insufficient parameters passed to orthknife utility; looking for <delete_knife_ind> <orth_view> <select_verts>" );
      return;
   }

   // Get params
   bool deleteKnifeInd = ( strcmp( argv[1], "0" ) != 0 );
   int view = atoi( argv[2] );
   bool selVerts = ( strcmp( argv[3], "0" ) != 0 );

   // Add undoable knife
   DDHUndoableOrthKnife *undoknife = new DDHUndoableOrthKnife( ddhStaticOrthKnife,
                                                               deleteKnifeInd,
                                                               view );
   add_undoable( "knife", undoknife );

   if( !selVerts )
   {
      return;
   }

   // Go into vertex mode
   List *newverts = undoknife->GetNewVerts();
   if( newverts == NULL )
   {
      // Nothing was done
      return;
   }
   if( ac_get_select_mode() != SELECT_VERTEX )
   {
      tcl_command( "ac3d set_select_mode 3" );
   }

   // Select the new vertices
   add_undoable_change_selection( "change selection" );
   clear_selection();
   ac_selection_select_vertexlist( newverts );
   list_free( &newverts );

   redraw_all();
   find_bounding_box();
   display_status();

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/orthknife.tcl\"", path );
      ac_add_tools_menu_item( "Ortho knife...", "ddhorthknife_dialog",
                              "Cut selected objects using a knife object's edges" );
      ac_add_command_full( "ddhorthknife_go", ( void * )ddhorthknife_selectedGo,
                            3, "argv", "<slice_surfaces_ind> <delete_knife_ind> <orth_view>",
                           "Knife (orth) utility" );
      ac_add_command( "ddhorthknife_set", ddhorthknife_setKnife );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Orth Knife - version 1.0 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Orth Knife utility - Slices the edges of surfaces using the edges of a knife object.  Intersections are determined based on an orthographic view (XY, XZ, YZ)" );
}
//---------------------------------------------------------------------------------------
